#ifndef ANANICY_CPP_VERSION_HPP
#define ANANICY_CPP_VERSION_HPP

#include <string_view>

#define MAJOR_VERSION 
#define MINOR_VERSION 
#define PATCH_VERSION 
static constexpr std::string_view VERSION = "";
static constexpr std::string_view COMMIT_HASH = "";
static constexpr std::string_view FULL_VERSION = "";
static constexpr std::string_view BUILD_TYPE = "None";

#endif // ANANICY_CPP_VERSION_HPP
